
# THIS FILE IS GENERATED FROM PYWAVELETS SETUP.PY
short_version = '1.0.1'
version = '1.0.1'
full_version = '1.0.1'
git_revision = 'Unknown'
release = True

if not release:
    version = full_version
