#  tests for pywavelets-1.0.1-py37h1d22016_0 (this is a generated file);
print('===== testing package: pywavelets-1.0.1-py37h1d22016_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import pywt
pywt.test()
#  --- run_test.py (end) ---

print('===== pywavelets-1.0.1-py37h1d22016_0 OK =====');
